import logging
from typing import Union, List

from . import _computers
from ._ldaps import LdapObject
from ._object_classes import Computer, Group

logger = logging.getLogger(__name__)


def search(connection: LdapObject, **search_params) -> List[Computer]:
    """ Search for an Active Directory computer.

    :param connection: An LDAPS connection.
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once automatically escaped,
        result in an invalid LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory computer objects.

    To find the computer with sAMAccountName 'test computer' , call the function like so:

    .. code-block:: python

        computers.search(connection, sAMAccountName="test computer ")

    Wildcards are supported provided they are at the end of the search string. To find all computers with a
    sAMAccountName starting with 'server-', call the function like so:

    .. code-block:: python

        computers.search(connection, sAMAccountName="server-*")

    .. note:: If the wildcard character is not at the end of the search string, it will be escaped.
    """
    logger.info(f"searching for computer with params {search_params}")
    return _computers._get(connection, **search_params)


def get_all(connection: LdapObject) -> List[Computer]:
    """ Get all Active Directory computers.

    :param connection: An LDAPS connection.

    :raises exceptions.UserAccountDisabled: if the computer account performing the operation is disabled.

    :returns: A list of Active Directory computer objects.

    Example:

        .. code-block:: python

            computers.get_all(connection)

    .. warning:: This operation can be expensive! Using the search function is recommended.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info("getting all computers")
    return _computers._get(connection)


def group_memberships(connection: LdapObject, *, distinguished_name: str) -> List[Group]:
    """ List the group memberships of a computer.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.

    :raises exceptions.NoSuchObject: if the computer does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: A list of Active Directory group objects.

    .. note:: This function does not find nested group memberships.

    .. seealso:: :ref:`active_directory:Search Functions`
    """
    logger.info(f"getting group memberships for computer with distinguished name {distinguished_name}")
    return _computers._group_memberships(connection, distinguished_name)


def exists(connection: LdapObject, **search_params) -> bool:
    """ Check whether a computer exists.

    :param connection: An LDAPS connection.
    :param search_params: Keyword arguments to include in the LDAP query.

    :raises exceptions.InvalidFilter: in the unlikely case that search parameters, once escaped, result in an invalid
        LDAP query.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :returns: Bool.

    Example:

        .. code-block:: python

            computers.exists(connection, sAMAccountName="test computer")

    .. seealso:: :ref:`active_directory:Boolean Functions`, :ref:`active_directory:Keyword Argument Attributes`
    """
    logger.info(f"checking computer exists with params {search_params}")
    return bool(search(connection, **search_params))


def set_attribute(connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]) -> None:
    """ Set a computer LDAP attribute value. This function performs the LDAP modify_add
    operation. It is suitable for setting values for blank attributes.

    If the supplied attribute already has a value, exceptions.ModifyAttributeError
    will be raised. In this case modify_attribute should be used instead.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to set on the LDAP attribute.

    :raises exceptions.ModifyAttributeError: if there was an issue setting the new attribute value.
    :raises exceptions.NoSuchObject: if the computer does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            computers.set_attribute(
                connection,
                distinguished_name="cn=test computer,cn=Computers,dc=test,dc=domain,dc=com",
                name="info",
                value="my new value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"setting attribute {name} to {value} on computer {distinguished_name}")
    _computers._modify(connection, distinguished_name, name, value, "add")


def replace_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Modify a computer LDAP attribute value. This function performs the LDAP modify_replace
    operation.

    As suggested by the function name; if the supplied attribute already has a value, it will
    be replaced.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The value to replace the existing LDAP attribute value with.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the computer does not exist.

    Example:

        .. code-block:: python

            computers.replace_attribute(
                connection,
                distinguished_name="cn=test computer,cn=Computers,dc=test,dc=domain,dc=com",
                name="info",
                value="my replaced value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"replacing attribute {name} to {value} on computer {distinguished_name}")
    _computers._modify(connection, distinguished_name, name, value, "replace")


def delete_attribute(
    connection: LdapObject, *, distinguished_name: str, name: str, value: Union[int, str, bytes]
) -> None:
    """ Remove a computer LDAP attribute value. This function performs the LDAP modify_delete
    operation.

    :param connection: An LDAPS connection.
    :param distinguished_name: The distinguished name of the target computer.
    :param attribute_name: The name of the LDAP attribute to modify.
    :param value: The attribute value to delete.

    :raises exceptions.ModifyAttributeError: if there was an issue replacing the attribute value.
    :raises exceptions.NoSuchObject: if the computer does not exist.
    :raises exceptions.UserAccountDisabled: if the user account performing the operation is disabled.

    Example:

        .. code-block:: python

            computers.delete_attribute(
                connection,
                distinguished_name="cn=test computer,cn=Computers,dc=test,dc=domain,dc=com",
                name="info",
                value="known existing value"
            )

    .. seealso:: :ref:`active_directory:Action Functions`
    """
    logger.info(f"deleting attribute {name} value {value} on computer {distinguished_name}")
    _computers._modify(connection, distinguished_name, name, value, "delete")
