import logging
import functools

from typing import Callable, Any

from active_directory import client, exceptions

from .records import DomainController

logger = logging.getLogger(__name__)


def idempotent_create(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_create(*args, **kwargs) -> None:
        # Idempotent create functions should not care if the object already exists.
        try:
            return func(*args, **kwargs)
        except exceptions.ObjectExists as e:
            logger.debug("idempotent create function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_create


def idempotent_remove(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_remove(*args, **kwargs) -> None:
        # Idempotent remove functions should not care if the object does not exist.
        try:
            return func(*args, **kwargs)
        except exceptions.NoSuchObject as e:
            logger.debug("idempotent remove function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_remove


def ldap_function(func: Callable) -> Callable:
    @functools.wraps(func)
    def _fn(domain_controller: DomainController, *args, **kwargs) -> Any:
        missing_keys = []
        for key in {"address", "domain", "username", "password"}:
            if key not in domain_controller:
                missing_keys.append(key)
        if "port" not in domain_controller:
            domain_controller["port"] = 636
        if missing_keys:
            raise Exception(
                "Active Directory functions require address, domain, username, and password in the "
                f"connection details. Received {list(domain_controller.keys())}."
            )
        with client.initialize(**domain_controller)() as open_connection:
            return func(open_connection, *args, **kwargs)

    return _fn
